@@ -3,37 +3,37 @@ Contains most of the loading logic.
 """
 
 import dataclasses as dc
-import math
-import types
+import io
+import os
 import typing
 import warnings
-from collections import ChainMap
 from pathlib import Path
+from typing import Any, Type
 
-from typeguard import TypeCheckError
-from typeguard import check_type as _check_type
+import requests
 
 from . import loaders
+from .abs import C, T, T_data, Type_C
+from .alias import Alias
+from .binary_config import BinaryConfig
 from .errors import (
     ConfigErrorCouldNotConvert,
     ConfigErrorInvalidType,
     ConfigErrorMissingKey,
 )
-from .helpers import camel_to_snake
+from .helpers import (
+    all_annotations,
+    camel_to_snake,
+    check_type,
+    dataclass_field,
+    find_pyproject_toml,
+    is_custom_class,
+    is_optional,
+    is_parameterized,
+)
 from .postpone import Postponed
 from .type_converters import CONVERTERS
 
-# T is a reusable typevar
-T = typing.TypeVar("T")
-# t_typelike is anything that can be type hinted
-T_typelike: typing.TypeAlias = type | types.UnionType  # | typing.Union
-# t_data is anything that can be fed to _load_data
-T_data = str | Path | dict[str, typing.Any]
-# c = a config class instance, can be any (user-defined) class
-C = typing.TypeVar("C")
-# type c is a config class
-Type_C = typing.Type[C]
-
 
 def _data_for_nested_key(key: str, raw: dict[str, typing.Any]) -> dict[str, typing.Any]:
     """
@@ -46,7 +46,11 @@ def _data_for_nested_key(key: str, raw: dict[str, typing.Any]) -> dict[str, typi
     """
     parts = key.split(".")
     while parts:
-        raw = raw[parts.pop(0)]
+        key = parts.pop(0)
+        if key not in raw:
+            return {}
+
+        raw = raw[key]
 
     return raw
 
@@ -59,8 +63,62 @@ def _guess_key(clsname: str) -> str:
     return camel_to_snake(clsname)
 
 
-def __load_data(
-    data: T_data, key: str = None, classname: str = None, lower_keys: bool = False
+def _from_mock_url(url: str) -> str:
+    """
+    Pytest only: when starting a url with mock:// it is expected to just be json afterwards.
+    """
+    return url.removeprefix("mock://")
+
+
+def guess_filetype_for_url(url: str, response: requests.Response = None) -> str:
+    """
+    Based on the url (which may have an extension) and the requests response \
+        (which may have a content-type), try to guess the right filetype (-> loader, e.g. json or yaml).
+
+    Falls back to JSON if none can be found.
+    """
+    url = url.split("?")[0]
+    if url_extension := os.path.splitext(url)[1].lower():
+        return url_extension.strip(".")
+
+    if response and (content_type_header := response.headers.get("content-type", "").split(";")[0].strip()):
+        content_type = content_type_header.split("/")[-1]
+        if content_type != "plain":
+            return content_type
+
+    # If both methods fail, default to JSON
+    return "json"
+
+
+def from_url(url: str, _dummy: bool = False) -> tuple[io.BytesIO, str]:
+    """
+    Load data as bytes into a file-like object and return the file type.
+
+    This can be used by __load_data:
+    > loader = loaders.get(filetype)
+    > # dev/null exists but always returns b''
+    > data = loader(contents, Path("/dev/null"))
+    """
+    if url.startswith("mock://"):
+        data = _from_mock_url(url)
+        resp = None
+    elif _dummy:
+        resp = None
+        data = "{}"
+    else:
+        resp = requests.get(url, timeout=10)
+        data = resp.text
+
+    filetype = guess_filetype_for_url(url, resp)
+    return io.BytesIO(data.encode()), filetype
+
+
+def _load_data(
+    data: T_data,
+    key: str = None,
+    classname: str = None,
+    lower_keys: bool = False,
+    allow_types: tuple[type, ...] = (dict,),
 ) -> dict[str, typing.Any]:
     """
     Tries to load the right data from a filename/path or dict, based on a manual key or a classname.
@@ -68,19 +126,44 @@ def __load_data(
     E.g. class Tool will be mapped to key tool.
     It also deals with nested keys (tool.extra -> {"tool": {"extra": ...}}
     """
+    if isinstance(data, bytes):
+        # instantly return, don't modify
+        # bytes as inputs -> bytes as output
+        # but since `T_data` is re-used, that's kind of hard to type for mypy.
+        return data  # type: ignore
+
+    if isinstance(data, list):
+        if not data:
+            raise ValueError("Empty list passed!")
+
+        final_data: dict[str, typing.Any] = {}
+        for source in data:
+            final_data |= load_data(source, key=key, classname=classname, lower_keys=True, allow_types=allow_types)
+
+        return final_data
+
     if isinstance(data, str):
-        data = Path(data)
+        if data.startswith(("http://", "https://", "mock://")):
+            contents, filetype = from_url(data)
+
+            loader = loaders.get(filetype)
+            # dev/null exists but always returns b''
+            data = loader(contents, Path("/dev/null"))
+        else:
+            data = Path(data)
+
     if isinstance(data, Path):
         with data.open("rb") as f:
             loader = loaders.get(data.suffix or data.name)
             data = loader(f, data.resolve())
+
     if not data:
         return {}
 
     if key is None:
         # try to guess key by grabbing the first one or using the class name
         if len(data) == 1:
-            key = list(data.keys())[0]
+            key = next(iter(data.keys()))
         elif classname is not None:
             key = _guess_key(classname)
 
@@ -90,24 +173,36 @@ def __load_data(
     if not data:
         raise ValueError("No data found!")
 
-    if not isinstance(data, dict):
-        raise ValueError("Data is not a dict!")
+    if not isinstance(data, allow_types):
+        raise ValueError(f"Data should be one of {allow_types} but it is {type(data)}!")
 
-    if lower_keys:
+    if lower_keys and isinstance(data, dict):
         data = {k.lower(): v for k, v in data.items()}
 
-    return data
+    return typing.cast(dict[str, typing.Any], data)
 
 
-def _load_data(data: T_data, key: str = None, classname: str = None, lower_keys: bool = False) -> dict[str, typing.Any]:
+def load_data(
+    data: T_data,
+    key: str = None,
+    classname: str = None,
+    lower_keys: bool = False,
+    allow_types: tuple[type, ...] = (dict,),
+) -> dict[str, typing.Any]:
     """
     Wrapper around __load_data that retries with key="" if anything goes wrong.
     """
+    if data is None:
+        # try to load pyproject.toml
+        data = find_pyproject_toml()
+
     try:
-        return __load_data(data, key, classname, lower_keys=lower_keys)
+        return _load_data(data, key, classname, lower_keys=lower_keys, allow_types=allow_types)
     except Exception as e:
+        # sourcery skip: remove-unnecessary-else, simplify-empty-collection-comparison, swap-if-else-branches
+        # @sourcery: `key != ""` is NOT the same as `not key`
         if key != "":
-            return __load_data(data, "", classname, lower_keys=lower_keys)
+            return _load_data(data, "", classname, lower_keys=lower_keys, allow_types=allow_types)
         else:  # pragma: no cover
             warnings.warn(f"Data could not be loaded: {e}", source=e)
             # key already was "", just return data!
@@ -115,23 +210,10 @@ def _load_data(data: T_data, key: str = None, classname: str = None, lower_keys:
             return {}
 
 
-def check_type(value: typing.Any, expected_type: T_typelike) -> bool:
-    """
-    Given a variable, check if it matches 'expected_type' (which can be a Union, parameterized generic etc.).
-
-    Based on typeguard but this returns a boolean instead of returning the value or throwing a TypeCheckError
-    """
-    try:
-        _check_type(value, expected_type)
-        return True
-    except TypeCheckError:
-        return False
-
-
 F = typing.TypeVar("F")
 
 
-def convert_between(from_value: F, from_type: typing.Type[F], to_type: type[T]) -> T:
+def convert_between(from_value: F, from_type: Type[F], to_type: Type[T]) -> T:
     """
     Convert a value between types.
     """
@@ -142,6 +224,46 @@ def convert_between(from_value: F, from_type: typing.Type[F], to_type: type[T])
     return to_type(from_value)  # type: ignore
 
 
+def check_and_convert_type(value: Any, _type: Type[T], convert_types: bool, key: str = "variable") -> T:
+    """
+    Checks if the given value matches the specified type. If it does, the value is returned as is.
+
+    Args:
+        value (Any): The value to be checked and potentially converted.
+        _type (Type[T]): The expected type for the value.
+        convert_types (bool): If True, allows type conversion if the types do not match.
+        key (str, optional): The name or key associated with the variable (used in error messages).
+                             Defaults to "variable".
+
+    Returns:
+        T: The value, potentially converted to the expected type.
+
+    Raises:
+        ConfigErrorInvalidType: If the type does not match, and type conversion is not allowed.
+        ConfigErrorCouldNotConvert: If type conversion fails.
+    """
+    if check_type(value, _type):
+        # type matches
+        return value
+
+    if isinstance(value, Alias):
+        if is_optional(_type):
+            return typing.cast(T, None)
+        else:
+            # unresolved alias, error should've already been thrown for parent but lets do it again:
+            raise ConfigErrorInvalidType(value.to, value=value, expected_type=_type)
+
+    if not convert_types:
+        # type does not match and should not be converted
+        raise ConfigErrorInvalidType(key, value=value, expected_type=_type)
+
+    # else: type does not match, try to convert it
+    try:
+        return convert_between(value, type(value), _type)
+    except (TypeError, ValueError) as e:
+        raise ConfigErrorCouldNotConvert(type(value), _type, value) from e
+
+
 def ensure_types(
     data: dict[str, T], annotations: dict[str, type[T]], convert_types: bool = False
 ) -> dict[str, T | None]:
@@ -155,7 +277,6 @@ def ensure_types(
     # custom object to use instead of None, since typing.Optional can be None!
     # cast to T to make mypy happy
     notfound = typing.cast(T, object())
-    postponed = Postponed()
 
     final: dict[str, T | None] = {}
     for key, _type in annotations.items():
@@ -167,18 +288,21 @@ def ensure_types(
             # skip!
             continue
 
-        if compare is postponed:
+        if isinstance(compare, Postponed):
             # don't do anything with this item!
             continue
 
-        if not check_type(compare, _type):
-            if convert_types:
-                try:
-                    compare = convert_between(compare, type(compare), _type)
-                except (TypeError, ValueError) as e:
-                    raise ConfigErrorCouldNotConvert(type(compare), _type, compare) from e
-            else:
-                raise ConfigErrorInvalidType(key, value=compare, expected_type=_type)
+        if isinstance(compare, Alias):
+            related_data = data.get(compare.to, notfound)
+            if related_data is not notfound:
+                if isinstance(related_data, Postponed):
+                    # also continue alias for postponed items
+                    continue
+
+                # original key set, update alias
+                compare = related_data
+
+        compare = check_and_convert_type(compare, _type, convert_types, key)
 
         final[key] = compare
 
@@ -195,105 +319,48 @@ def convert_config(items: dict[str, T]) -> dict[str, T]:
     return {k.replace("-", "_").replace(".", "_"): v for k, v in items.items() if v is not None}
 
 
-Type = typing.Type[typing.Any]
-T_Type = typing.TypeVar("T_Type", bound=Type)
+AnyType: typing.TypeAlias = typing.Type[typing.Any]
+T_Type = typing.TypeVar("T_Type", bound=AnyType)
 
 
-def is_builtin_type(_type: Type) -> bool:
+def has_aliases(cls: AnyType, key: str) -> typing.Generator[str, None, None]:
     """
-    Returns whether _type is one of the builtin types.
+    Generate all aliases that point to 'key' in 'cls'.
     """
-    return _type.__module__ in ("__builtin__", "builtins")
+    for field, value in cls.__dict__.items():
+        if isinstance(value, Alias) and value.to == key:
+            yield field
 
 
-# def is_builtin_class_instance(obj: typing.Any) -> bool:
-#     return is_builtin_type(obj.__class__)
-
-
-def is_from_types_or_typing(_type: Type) -> bool:
+def has_alias(cls: AnyType, key: str, data: dict[str, T]) -> typing.Optional[T]:
     """
-    Returns whether _type is one of the stlib typing/types types.
+    Get the value of any alias in the same config class that references `key`.
 
-    e.g. types.UnionType or typing.Union
+    Example:
+        class Config:
+            key1: str
+            key2: str = alias('key1')
+
+    load_into(Config, {'key2': 'something'})
+    # -> key1 will look up the value of key2 because it's configured as an alias for it.
+
+    If multiple aliases point to the same base, they are all iterated until a valid value was found.
     """
-    return _type.__module__ in ("types", "typing")
+    # for field, value in cls.__dict__.items():
+    #     if isinstance(value, Alias) and value.to == key:
+    #         # yay!
+    #         return data.get(field)
+    #
+    # return None
 
-
-def is_from_other_toml_supported_module(_type: Type) -> bool:
-    """
-    Besides builtins, toml also supports 'datetime' and 'math' types, \
-    so this returns whether _type is a type from these stdlib modules.
-    """
-    return _type.__module__ in ("datetime", "math")
-
-
-def is_parameterized(_type: Type) -> bool:
-    """
-    Returns whether _type is a parameterized type.
-
-    Examples:
-        list[str] -> True
-        str -> False
-    """
-    return typing.get_origin(_type) is not None
-
-
-def is_custom_class(_type: Type) -> bool:
-    """
-    Tries to guess if _type is a builtin or a custom (user-defined) class.
-
-    Other logic in this module depends on knowing that.
-    """
-    return (
-        type(_type) is type
-        and not is_builtin_type(_type)
-        and not is_from_other_toml_supported_module(_type)
-        and not is_from_types_or_typing(_type)
+    return next(
+        (value for field in has_aliases(cls, key) if (value := data.get(field))),
+        None,
     )
 
 
-def instance_of_custom_class(var: typing.Any) -> bool:
-    """
-    Calls `is_custom_class` on an instance of a (possibly custom) class.
-    """
-    return is_custom_class(var.__class__)
-
-
-def is_optional(_type: Type | typing.Any) -> bool:
-    """
-    Tries to guess if _type could be optional.
-
-    Examples:
-        None -> True
-        NoneType -> True
-        typing.Union[str, None] -> True
-        str | None -> True
-        list[str | None] -> False
-        list[str] -> False
-    """
-    if _type and (is_parameterized(_type) and typing.get_origin(_type) in (dict, list)) or (_type is math.nan):
-        # e.g. list[str]
-        # will crash issubclass to test it first here
-        return False
-
-    return (
-        _type is None
-        or types.NoneType in typing.get_args(_type)  # union with Nonetype
-        or issubclass(types.NoneType, _type)
-        or issubclass(types.NoneType, type(_type))  # no type  # Nonetype
-    )
-
-
-def dataclass_field(cls: Type, key: str) -> typing.Optional[dc.Field[typing.Any]]:
-    """
-    Get Field info for a dataclass cls.
-    """
-    fields = getattr(cls, "__dataclass_fields__", {})
-    return fields.get(key)
-
-
 def load_recursive(
-    cls: Type, data: dict[str, T], annotations: dict[str, Type], convert_types: bool = False
+    cls: AnyType, data: dict[str, T], annotations: dict[str, AnyType], convert_types: bool = False
 ) -> dict[str, T]:
     """
     For all annotations (recursively gathered from parents with `all_annotations`), \
@@ -351,9 +418,6 @@ def load_recursive(
                     for arg in arguments:
                         if is_custom_class(arg):
                             value = _load_into_recurse(arg, value, convert_types=convert_types)
-                        else:
-                            # print(_type, arg, value)
-                            ...
 
                 # todo: other parameterized/unions/typing.Optional
 
@@ -366,7 +430,9 @@ def load_recursive(
                     value,
                     convert_types=convert_types,
                 )
-
+        elif value := has_alias(cls, _key, data):
+            # value updated by alias
+            ...
         elif _key in cls.__dict__:
             # property has default, use that instead.
             value = cls.__dict__[_key]
@@ -385,27 +451,6 @@ def load_recursive(
     return updated
 
 
-def _all_annotations(cls: Type) -> ChainMap[str, Type]:
-    """
-    Returns a dictionary-like ChainMap that includes annotations for all \
-    attributes defined in cls or inherited from superclasses.
-    """
-    return ChainMap(*(c.__annotations__ for c in getattr(cls, "__mro__", []) if "__annotations__" in c.__dict__))
-
-
-def all_annotations(cls: Type, _except: typing.Iterable[str] = None) -> dict[str, type[object]]:
-    """
-    Wrapper around `_all_annotations` that filters away any keys in _except.
-
-    It also flattens the ChainMap to a regular dict.
-    """
-    if _except is None:
-        _except = set()
-
-    _all = _all_annotations(cls)
-    return {k: v for k, v in _all.items() if k not in _except}
-
-
 def check_and_convert_data(
     cls: typing.Type[C],
     data: dict[str, typing.Any],
@@ -424,6 +469,7 @@ def check_and_convert_data(
 
     to_load = convert_config(data)
     to_load = load_recursive(cls, to_load, annotations, convert_types=convert_types)
+
     if strict:
         to_load = ensure_types(to_load, annotations, convert_types=convert_types)
 
@@ -458,7 +504,7 @@ def _split_init(init: T_init) -> tuple[T_init_list, T_init_dict]:
 
 def _load_into_recurse(
     cls: typing.Type[C],
-    data: dict[str, typing.Any],
+    data: dict[str, typing.Any] | bytes,
     init: T_init = None,
     strict: bool = True,
     convert_types: bool = False,
@@ -472,7 +518,14 @@ def _load_into_recurse(
     """
     init_args, init_kwargs = _split_init(init)
 
-    if dc.is_dataclass(cls):
+    if isinstance(data, bytes) or issubclass(cls, BinaryConfig):
+        if not isinstance(data, (bytes, dict)):  # pragma: no cover
+            raise NotImplementedError("BinaryConfig can only deal with `bytes` or a dict of bytes as input.")
+        elif not issubclass(cls, BinaryConfig):  # pragma: no cover
+            raise NotImplementedError("Only BinaryConfig can be used with `bytes` (or a dict of bytes) as input.")
+
+        inst = typing.cast(C, cls._parse_into(data))
+    elif dc.is_dataclass(cls):
         to_load = check_and_convert_data(cls, data, init_kwargs.keys(), strict=strict, convert_types=convert_types)
         if init:
             raise ValueError("Init is not allowed for dataclasses!")
@@ -527,7 +580,8 @@ def load_into_class(
     """
     Shortcut for _load_data + load_into_recurse.
     """
-    to_load = _load_data(data, key, cls.__name__, lower_keys=lower_keys)
+    allow_types = (dict, bytes) if issubclass(cls, BinaryConfig) else (dict,)
+    to_load = load_data(data, key, cls.__name__, lower_keys=lower_keys, allow_types=allow_types)
     return _load_into_recurse(cls, to_load, init=init, strict=strict, convert_types=convert_types)
 
 
@@ -545,13 +599,14 @@ def load_into_instance(
     Shortcut for _load_data + load_into_existing.
     """
     cls = inst.__class__
-    to_load = _load_data(data, key, cls.__name__, lower_keys=lower_keys)
+    allow_types = (dict, bytes) if issubclass(cls, BinaryConfig) else (dict,)
+    to_load = load_data(data, key, cls.__name__, lower_keys=lower_keys, allow_types=allow_types)
     return _load_into_instance(inst, cls, to_load, init=init, strict=strict, convert_types=convert_types)
 
 
 def load_into(
     cls: typing.Type[C],
-    data: T_data,
+    data: T_data = None,
     /,
     key: str = None,
     init: T_init = None,
