@@ -1,12 +1,14 @@
 # not directly testable
+import io
 import math
-import sys
 import typing
+from pathlib import Path
 
 import pytest
 
 from src.configuraptor import all_annotations
 from src.configuraptor.core import is_optional
+from src.configuraptor.helpers import as_binaryio
 from src.configuraptor.type_converters import str_to_none
 
 
@@ -22,6 +24,26 @@ def test_is_optional_with_weird_inputs():
     assert is_optional(typing.Optional[dict[str, typing.Optional[str]]]) is True
 
 
+def test_as_binaryio():
+    path = Path("/tmp/pytest_asbinary_file")
+    path.touch()
+
+    with as_binaryio(str(path)) as f:
+        assert hasattr(f, "read")
+
+    with as_binaryio(path) as f:
+        assert hasattr(f, "read")
+
+    with as_binaryio(open(str(path), "rb")) as f:
+        assert hasattr(f, "read")
+
+    with as_binaryio(io.BytesIO()) as f:
+        assert hasattr(f, "read")
+
+    with as_binaryio(None) as f:
+        assert hasattr(f, "read")
+
+
 class Base:
     has: int
 
@@ -42,17 +64,17 @@ def test_no_data():
 
     # data must be a dict:
     with pytest.raises(ValueError):
-        configuraptor.core._load_data(42, key=None)
+        configuraptor.core.load_data(42, key=None)
     with pytest.raises(ValueError):
-        configuraptor.core._load_data(["joe"], key=None)
+        configuraptor.core.load_data([], key=None)
 
     # but other than that, it should be fine:
-    configuraptor.core._load_data({}, key="")
-    configuraptor.core._load_data({}, key=None)
-    configuraptor.core._load_data({"-": 0, "+": None}, key="joe", classname="-.+")
-    configuraptor.core._load_data({"-": 0, "+": None}, key="+", classname="-.+")
-    configuraptor.core._load_data({"-": 0, "+": None}, key="", classname="-.+")
-    configuraptor.core._load_data({"-": 0, "+": None}, key=None, classname="-.+")
+    configuraptor.core.load_data({}, key="")
+    configuraptor.core.load_data({}, key=None)
+    configuraptor.core.load_data({"-": 0, "+": None}, key="joe", classname="-.+")
+    configuraptor.core.load_data({"-": 0, "+": None}, key="+", classname="-.+")
+    configuraptor.core.load_data({"-": 0, "+": None}, key="", classname="-.+")
+    configuraptor.core.load_data({"-": 0, "+": None}, key=None, classname="-.+")
 
 
 def test_str_to_none():
